/* Copyright (c) 2010-2011, Freescale Semiconductor, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Freescale Semiconductor nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 *
 * ALTERNATIVELY, this software may be distributed under the terms of the
 * GNU General Public License ("GPL") as published by the Free Software
 * Foundation, either version 2 of that License or (at your option) any
 * later version.
 *
 * THIS SOFTWARE IS PROVIDED BY Freescale Semiconductor ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL Freescale Semiconductor BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <linux/skbuff.h>
#include <linux/in.h>
#include <linux/ip.h>
#include <linux/udp.h>
#include <linux/tcp.h>
#include <linux/delay.h>
#include <linux/jiffies.h>
#include <linux/fs.h>
#include <linux/proc_fs.h>
#include <linux/netfilter.h>
#include <linux/netfilter_ipv4.h>
#include <linux/fsl_pme.h>
#include <linux/workqueue.h>
#include <asm/time.h>
#include <linux/module.h>
#include <linux/seq_file.h>
#include <linux/cpu.h>
#include <linux/kernel.h>
#include <linux/of.h>

#define MAX_CORES	8
#define MAX_CONTEXT_PER_IF	(5)
#define MAX_INTERFACE	10

/* basic requirement check */
#ifndef CONFIG_NETFILTER
#error "Error: NETFILTER is not enabled in the kernel"
#endif

MODULE_AUTHOR("Freescale Semiconductor, Inc");
MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("PME network scanning example");

static int num_cores;
module_param(num_cores, int, S_IRUGO);
MODULE_PARM_DESC(num_cores, "Number of cores, default=num online");

static int use_orp;
module_param(use_orp, int, S_IRUGO);
MODULE_PARM_DESC(use_orp, "Use order restoration, default=0");

static int flow_mode = 1;
module_param(flow_mode, int, S_IRUGO);
MODULE_PARM_DESC(flow_mode, "Use flow mode, default=1");

static int scan_header;
module_param(scan_header, int, S_IRUGO);
MODULE_PARM_DESC(scan_header, "Defines if the packet header is part of"
				" the scan, default=0");
static int skip_pme;
module_param(skip_pme, int, S_IRUGO);
MODULE_PARM_DESC(skip_pme, "Hook into netfilter, but return immediatly");

static int qos_in;
module_param(qos_in, int, S_IRUGO);
MODULE_PARM_DESC(qos_in, "qos_in level default=0");

static int qos_out;
module_param(qos_out, int, S_IRUGO);
MODULE_PARM_DESC(qos_out, "qos_out level default=0");

static int use_stash;
module_param(use_stash, int, S_IRUGO);
MODULE_PARM_DESC(use_stash, "Use PME stashing");

static int stash_exc;
module_param(stash_exc, int, S_IRUGO);
MODULE_PARM_DESC(stash_exc, "Stashing, exce, ANN=0x4, DATA=0x2, CTX=0x1");

static int stash_annotation_cl;
module_param(stash_annotation_cl, int, S_IRUGO);
MODULE_PARM_DESC(stash_annotation_cl, "Stashing, annotation (0..3)");

static int stash_data_cl;
module_param(stash_data_cl, int, S_IRUGO);
MODULE_PARM_DESC(stash_data_cl, "Stashing, data_cl (0..3)");

static int stash_ctx_cl;
module_param(stash_ctx_cl, int, S_IRUGO);
MODULE_PARM_DESC(stash_ctx_cl, "Stashing, context_cl (0..3)");

static int fm1_mask = 0xC; /* fm1-gb2, fm1-gb3 */
module_param(fm1_mask, int, S_IRUGO);
MODULE_PARM_DESC(fm1_mask, "Fm1 bit mask, which ethernet devices to scan"
" pme data=0. For lower bits are fm1-gb[0..3]. Fifth bit is fm1-10g.");

static int fm2_mask = 0xC;
module_param(fm2_mask, int, S_IRUGO);
MODULE_PARM_DESC(fm2_mask, "Fm2 bit mask, which ethernet devices to scan"
" pme data=0. For lower bits are fm2-gb[0..3]. Fifth bit is fm2-10g.");

static int num_ctx_per_if = 5;
module_param(num_ctx_per_if, int, S_IRUGO);
MODULE_PARM_DESC(num_ctx_per_if, "Number of pme ctx per interface."
				" default=5");

static int allocs = 1000;
module_param(allocs, int, S_IRUGO);
MODULE_PARM_DESC(allocs, "Number of kmallocs outstanding");

static int residue;
module_param(residue, int, S_IRUGO);
MODULE_PARM_DESC(residue, "Residue enabled or not default=0");

static unsigned int active_cores;
static unsigned long start_time;
static unsigned int num_interfaces;
static int interface_active[MAX_INTERFACE];
static int qman_channel_config[MAX_INTERFACE];
static int interface_being_used[MAX_INTERFACE];
static atomic_t tokens_outstanding;

enum pme_hw_item {
	INPUT_BYTES,
	ONE_BYTE_TRIGGER_HITS,
	TWO_BYTE_TRIGGER_HITS,
	VAR_BYTE_TRIGGER_HITS,
	SPECIAL_ATRIGGER_HITS,
	CONFIDENCE_HITS,
	HW_MATCHES,
	SUIS_WITH_MATCHES,
	INPUT_SUIS,
	MIA_LS1_HITS,
	MIA_CW1,
	NR_PME_HW_ITEM
};

static u64 hw_stats[NR_PME_HW_ITEM];

/* Per cpu  event stats*/
enum pme_stat_item {
	RX_PACKETS,
	IP_PACKETS,
	DATA_PACKETS,
	FULL_FIFO,
	MATCHES,
	TOTAL_MATCHES,
	ERRORED_NOTIFS,
	TOTAL_NOTIFS,
	STACK_DROPS,
	NUM_TRUNCATES,
	NUM_WITH_OUTPUT,
	NUM_QUEUE_EMPTY,
	NUM_ERNS,
	NO_SCANS,
	SCAN_IN_IRQ,
	NR_PME_EVENTS
};

struct pme_stats {
	u64 counter[NR_PME_EVENTS];
};

static DEFINE_PER_CPU(struct pme_stats, pme_stat_counts);

static inline void __pme_stat_inc(enum pme_stat_item item)
{
	__get_cpu_var(pme_stat_counts).counter[item]++;
}

static inline void pme_stat_inc(enum pme_stat_item item)
{
	get_cpu_var(pme_stat_counts).counter[item]++;
	put_cpu();
}

static inline void __pme_stat_clear(enum pme_stat_item item)
{
	__get_cpu_var(pme_stat_counts).counter[item] = 0;
}

static inline void pme_stat_clear(enum pme_stat_item item)
{
	get_cpu_var(pme_stat_counts).counter[item] = 0;
	put_cpu();
}

static void __clear_pme_stats(const struct cpumask *cpumask)
{
	int cpu;

	for_each_cpu(cpu, cpumask) {
		struct pme_stats *this = &per_cpu(pme_stat_counts, cpu);
		memset(this, 0, sizeof(*this));
	}
}

static void clear_all_pme_stats(void)
{
	get_online_cpus();
	__clear_pme_stats(cpu_online_mask);
	put_online_cpus();
}


static void sum_pme_stats(u64 *ret, const struct cpumask *cpumask)
{
	int cpu;
	int i;

	memset(ret, 0, NR_PME_EVENTS * sizeof(u64));

	for_each_cpu(cpu, cpumask) {
		struct pme_stats *this = &per_cpu(pme_stat_counts, cpu);

		for (i = 0; i < NR_VM_EVENT_ITEMS; i++)
			ret[i] += this->counter[i];
	}
}

static void all_pme_stats(u64 *ret)
{
	get_online_cpus();
	sum_pme_stats(ret, cpu_online_mask);
	put_online_cpus();
}

struct cb_data {
	struct list_head	list;
	spinlock_t		list_lock;
	struct tasklet_struct	completion_tasklet;
};

struct ctrl_op {
	struct pme_ctx_ctrl_token ctx_ctr;
	struct completion cb_done;
	enum pme_status cmd_status;
	u8 res_flag;
};

struct skbuf_token {
	struct pme_ctx_token ctx_token;
	struct sk_buff *skb;
	int (*okfn)(struct sk_buff *);
	struct qm_fd resp_fd;
	int ern;
	/* tokens are added to cb_data list for delayed processing */
	struct list_head list;
};

static struct pme_ctx ctx_array[MAX_INTERFACE][MAX_CONTEXT_PER_IF];
/* first 4, fm1_[0..3], nex//t four, fm2_[0..3] */
static int interface_mapping[MAX_INTERFACE];
static struct cb_data defer_work[MAX_CORES];

static inline int is_interface_enabled(struct sk_buff *skb)
{
	int offset = 0;
	/* fm1/2 */
	if (skb->dev->name[2] == '2')
		offset = (MAX_INTERFACE/2);
	if (unlikely(skb->dev->name[6] == 'g'))
		return interface_active[offset + 4];
	else
		return interface_active[(skb->dev->name[6] - '0') + offset];
}

static inline unsigned which_interface(struct sk_buff *skb)
{
	int offset = 0;
	unsigned ret;

	if (skb->dev->name[2] == '2')
		offset = (MAX_INTERFACE/2);
	if (unlikely(skb->dev->name[6] == 'g'))
		ret = offset + 4;
	else
		ret = (skb->dev->name[6] - '0') + offset;
	return interface_mapping[ret];
}

/* There can be 256 flows per interface */
static inline u16 which_flow(struct sk_buff *skb)
{
	struct iphdr *iph = (struct iphdr *) skb->data;
	return (iph->saddr & 0xffff) % num_ctx_per_if;
}

/* Reads statistics from the PME HW */
static void get_stats(int reset)
{
	pme_stat_get(pme_attr_stnib, &hw_stats[INPUT_BYTES], reset);
	pme_stat_get(pme_attr_stnth1, &hw_stats[ONE_BYTE_TRIGGER_HITS],
			reset);
	pme_stat_get(pme_attr_stnth2, &hw_stats[TWO_BYTE_TRIGGER_HITS],
			reset);
	pme_stat_get(pme_attr_stnthv, &hw_stats[VAR_BYTE_TRIGGER_HITS],
			reset);
	pme_stat_get(pme_attr_stnths, &hw_stats[SPECIAL_ATRIGGER_HITS],
			reset);
	pme_stat_get(pme_attr_stnch, &hw_stats[CONFIDENCE_HITS], reset);
	pme_stat_get(pme_attr_stnpm, &hw_stats[HW_MATCHES], reset);
	pme_stat_get(pme_attr_stns1m, &hw_stats[SUIS_WITH_MATCHES], reset);
	pme_stat_get(pme_attr_stnis, &hw_stats[INPUT_SUIS], reset);
	pme_stat_get(pme_attr_mia_byc, &hw_stats[MIA_LS1_HITS], reset);
	pme_stat_get(pme_attr_mia_blc, &hw_stats[MIA_CW1], reset);
}

/* Determines if a packet should be scanned */
static inline int should_scan(struct sk_buff* skb) {
	struct iphdr* iph;

	if (!is_interface_enabled(skb))
		return 0;

	/* Check if this is an IP packet */
	if (unlikely(skb->protocol != 0x800))
		return 0;

	__pme_stat_inc(IP_PACKETS);

	iph = (struct iphdr*) skb->data;
	if (unlikely(iph->protocol != IPPROTO_TCP
			&& iph->protocol != IPPROTO_UDP))
		return 0;
	return 1;
}


/* Return the scanning context based on classification */
/* Classify based on which interface the packet was received on */
/* Classify based on last byte of the source address */
static inline struct pme_ctx *classify_packet(struct sk_buff* skb)
{
	return &ctx_array[which_interface(skb)][which_flow(skb)];
}

/*
 * These functions are used to provide a /proc interface to statistics
 * They aren't critical to the scanning demo and not part of the fast path
 */
static void* stat_start(struct seq_file *sfile, loff_t *pos)
{
	if (*pos != 0)
		return NULL;
	return pos;
}

static void* stat_next(struct seq_file *sfile, void* v, loff_t *pos)
{
	++(*pos);
	return NULL;
}

static void stat_stop(struct seq_file *s, void* v)
{
	return;
}

static int stat_show(struct seq_file *s, void *v)
{
	unsigned long end_time;
	u64 percent;

	u64 counters[NR_PME_EVENTS];

	all_pme_stats(counters);

	end_time = jiffies;

	seq_printf(s, "Processed %llu packets \n"
		"\t(%llu IP packets)"
		"\t (%llu TCP/UDP packets)\n",
		counters[RX_PACKETS],
		counters[IP_PACKETS],
		counters[DATA_PACKETS]);
	seq_printf(s, "%llu packets matched (%llu total matches)\n",
		counters[MATCHES], counters[TOTAL_MATCHES]);

	seq_printf(s, "Test ran for %lu msecs\n",
		((end_time - start_time) * 1000)/HZ);

	percent = counters[TOTAL_NOTIFS];
	do_div(percent, (end_time - start_time)/HZ);
	seq_printf(s, "Got %llu notifs, %llu notifs per sec\n",
		counters[TOTAL_NOTIFS], percent);
	//rate = percent * 8 * buffer_size;
	//do_div(rate, 1000 * 1000 );
	//seq_printf(s, "Bandwidth is %llu Mbps\n", rate);

	//seq_printf(s, "RAW DATA:\n\"%d\",\"%llu\",\"%d\"\n", buffer_size, percent, num_threads);

	seq_printf(s, "%llu truncates\n", counters[NUM_TRUNCATES]);
	seq_printf(s, "Scans with result data: %llu\n",
		counters[NUM_WITH_OUTPUT]);

	percent = counters[RX_PACKETS];
	do_div(percent, (end_time - start_time)/HZ);
	seq_printf(s, "Receiving %llu packets per sec\n", percent);
	seq_printf(s, "Frame queues were empty %llu times\n",
		counters[NUM_QUEUE_EMPTY]);
	seq_printf(s, "Full fifo	               : %llu\n",
		counters[FULL_FIFO]);
	seq_printf(s, "Num ERNS                 (ERN) : %llu\n",
		counters[NUM_ERNS]);
	seq_printf(s, "Num No Scans                    : %llu\n",
		counters[NO_SCANS]);


	if (counters[DATA_PACKETS]) {
		percent = counters[FULL_FIFO] * 100;
		do_div(percent, counters[DATA_PACKETS]);
		seq_printf(s, "%llu weren't scanned because the EQCR was"
			" full (%llu%%)\n", counters[FULL_FIFO], percent);
	}
	get_stats(0);
	seq_printf(s, "PM Input Bytes           (KES) : %llu\n",
		hw_stats[INPUT_BYTES]);
	seq_printf(s, "PM Trigger 1B Hits       (KES) : %llu\n",
		hw_stats[ONE_BYTE_TRIGGER_HITS]);
	seq_printf(s, "PM Trigger 2B Hits       (KES) : %llu\n",
		hw_stats[TWO_BYTE_TRIGGER_HITS]);
	seq_printf(s, "PM Trigger Variable Hits (KES) : %llu\n",
		hw_stats[VAR_BYTE_TRIGGER_HITS]);
	seq_printf(s, "PM Trigger Special Hits  (KES) : %llu\n",
		hw_stats[SPECIAL_ATRIGGER_HITS]);
	seq_printf(s, "PM Confidence Stage Hits (KES) : %llu\n",
		hw_stats[CONFIDENCE_HITS]);
	seq_printf(s, "PM Matches               (DXE) : %llu\n",
		hw_stats[HW_MATCHES]);
	seq_printf(s, "PM SUI With Matches      (DXE) : %llu\n",
		hw_stats[SUIS_WITH_MATCHES]);
	seq_printf(s, "PM Input SUIs            (KES) : %llu\n",
		hw_stats[INPUT_SUIS]);
	seq_printf(s, "Local Snoop 1 Hits       (MIA) : %llu\n",
		hw_stats[MIA_LS1_HITS]);
	seq_printf(s, "Cancelled Writes 1       (MIA) : %llu\n",
		hw_stats[MIA_CW1]);

	seq_printf(s, "tokens outstanding : %d\n",
		atomic_read(&tokens_outstanding));

	//seq_printf (s,"\nDCP 3 DLM_AVG=0x%x\n", qman_get_reg(0x33c));
	//seq_printf (s,"\nCI_RLM_AVG=0x%x\n", qman_get_reg(0xD14));
	//seq_printf (s,"\nPFDR_FPC=0x%x\n", qman_get_reg(0x400));
	return 0;
}

static ssize_t stat_write(struct file *filp, const char __user *buff,
				size_t len, loff_t *data)
{
	pr_info("Clearing net_scan_demo stats\n");
	memset(hw_stats, 0, sizeof(hw_stats));
	get_stats(1);
	clear_all_pme_stats();
	start_time = jiffies;
	return len;
}

static struct seq_operations stat_seq_ops =
{
	.start = stat_start,
	.next = stat_next,
	.stop = stat_stop,
	.show = stat_show
};

static int stat_proc_open(struct inode *inode, struct file* file)
{
	return seq_open(file, &stat_seq_ops);
}

static struct file_operations proc_ops =
{
	.owner = THIS_MODULE,
	.open = stat_proc_open,
	.read = seq_read,
	.write = stat_write,
	.llseek = seq_lseek,
	.release = seq_release
};

#ifdef DQRR_CONSUME
static int ctrl_cb(struct pme_ctx *ctx, const struct qm_fd *fd,
		struct pme_ctx_ctrl_token *token)
#else
static void ctrl_cb(struct pme_ctx *ctx, const struct qm_fd *fd,
		struct pme_ctx_ctrl_token *token)
#endif
{
	struct ctrl_op *ctrl = (struct ctrl_op *)token;

	ctrl->cmd_status = pme_fd_res_status(fd);
	ctrl->res_flag = pme_fd_res_flags(fd);
	complete(&ctrl->cb_done);
#ifdef DQRR_CONSUME
	return qman_cb_dqrr_consume;
#endif
};

static unsigned int packet_intercept(unsigned int hooknum,
		struct sk_buff *skb, const struct net_device *in,
		const struct net_device *out, int (*okfn)(struct sk_buff *))
{
	struct iphdr* iph;
	int ret;
	struct skbuf_token *sk_token = NULL;
	struct pme_ctx *ctx = NULL;
	struct qm_fd fd;
	dma_addr_t dma_addr;
	struct pme_ctx_token* token;

	__pme_stat_inc(RX_PACKETS);

	/* see if we need to scan the packet */
	if (unlikely(!should_scan(skb) || skip_pme)) {
		__pme_stat_inc(NO_SCANS);
		goto no_scan;
	}

	ctx = classify_packet(skb);
	if (atomic_dec_and_test(&tokens_outstanding)) {
		if (net_ratelimit())
			pr_err("Token allocation exceeded\n");
		atomic_inc(&tokens_outstanding);
		return NF_DROP;
	}
	sk_token = kmalloc(sizeof(*sk_token), GFP_ATOMIC);
	if (unlikely(!sk_token)) {
		atomic_inc(&tokens_outstanding);
		pr_err("kmalloc failed\n");
		return NF_DROP;
	}
	__pme_stat_inc(DATA_PACKETS);
	INIT_LIST_HEAD(&sk_token->list);
	token = &sk_token->ctx_token;
	memset(&fd, 0, sizeof(fd));
	iph = (struct iphdr*) skb->data;

	if (unlikely(!scan_header)) {
		/* Calculate the start point of the scan */
		if (iph->protocol == IPPROTO_TCP) {
			if (iph->tot_len <= (sizeof(struct tcphdr) +
				sizeof(struct iphdr)))
				goto no_scan;
			fd.length29 = iph->tot_len - sizeof(struct tcphdr) -
				sizeof(struct iphdr);
			dma_addr = pme_map(skb->data) +
				sizeof(struct tcphdr) +
				sizeof(struct iphdr);
		} else {
			/* UDP */
			if (iph->tot_len <= (sizeof(struct udphdr) +
				sizeof(struct iphdr)))
				goto no_scan;
			fd.length29 = iph->tot_len - sizeof(struct udphdr) -
				sizeof(struct iphdr);
			dma_addr = pme_map(skb->data) +
				sizeof(struct udphdr) +
				sizeof(struct iphdr);
		}
	} else {
		fd.length29 = iph->tot_len;
		dma_addr = pme_map(skb->data);
	}
	fd.format = qm_fd_contig_big;
	fd.addr_hi = upper_32_bits(dma_addr);
	fd.addr_lo = lower_32_bits(dma_addr);

	/* We don't verify checksums in this example.
	 * If the packet looks suspicious we skip it */
	if (unlikely(fd.length29 > 1500))
		goto no_scan;

	/* Start the scan */
	/* Need to save the okfn and skb */
	sk_token->okfn = okfn;
	sk_token->skb = skb;

	ret = pme_ctx_scan(ctx, 0, &fd,
		PME_SCAN_ARGS(0, 0, 0xff00),
		&sk_token->ctx_token);
	if (unlikely(ret)) {
		/* We're dropping this packet */
		if (sk_token)
			atomic_inc(&tokens_outstanding);
		kfree(sk_token);
		skb->sk = NULL;
		if (ret == -EBUSY)
			__pme_stat_inc(FULL_FIFO);
		else
			pr_err("pme_ctx_scan failed %d\n", ret);
		return NF_DROP;
	}
	
	return NF_STOLEN;
no_scan:
	if (sk_token)
		atomic_inc(&tokens_outstanding);
	kfree(sk_token);
	return NF_ACCEPT;
}

/* Netfilter hook */
static struct nf_hook_ops myhook_ops = {
	.hook = packet_intercept,
	.pf = PF_INET,
	.hooknum = NF_INET_PRE_ROUTING,
	.priority = NF_IP_PRI_FIRST ,
};

static void process_scanned_tokens(unsigned long _data)
{
	struct skbuf_token	*sk_token;
	struct sk_buff* skb;
	int (*okfn)(struct sk_buff *);
	enum pme_status pm_status;
	unsigned long fd_flags;
	unsigned int coreid = smp_processor_id();
	struct list_head *queue = &defer_work[coreid].list;
	unsigned long flags;

	spin_lock_irqsave(&defer_work[coreid].list_lock, flags);
	while (!list_empty(queue))  {
		sk_token = list_first_entry(queue, struct skbuf_token,
					list);
		list_del(&sk_token->list);
		spin_unlock_irqrestore(&defer_work[coreid].list_lock,
					flags);
		okfn = sk_token->okfn;
		skb = sk_token->skb;

		if (unlikely(sk_token->ern))
			goto token_free;

		pm_status = pme_fd_res_status(&sk_token->resp_fd);
		fd_flags = pme_fd_res_flags(&sk_token->resp_fd);

		if (unlikely(pm_status != pme_status_ok))
			pr_err("PME error status 0x%x\n", pm_status);
		/* Is truncation */
		if (unlikely(fd_flags & PME_STATUS_TRUNCATED))
			__pme_stat_inc(NUM_TRUNCATES);
token_free:
		atomic_inc(&tokens_outstanding);
		kfree(sk_token);
		okfn(skb);
		spin_lock_irqsave(&defer_work[coreid].list_lock, flags);
	}
	spin_unlock_irqrestore(&defer_work[coreid].list_lock, flags);
}

/* This function is invoked when a scan is completed */
#ifdef DQRR_CONSUME
static int scan_cb(struct pme_ctx *ctx, const struct qm_fd *fd,
		struct pme_ctx_token *token)
#else
static void scan_cb(struct pme_ctx *ctx, const struct qm_fd *fd,
		struct pme_ctx_token *token)
#endif
{
	struct skbuf_token* sk_token = (struct skbuf_token*)token;
	unsigned int coreid = smp_processor_id();
	unsigned long flags;
	int (*okfn)(struct sk_buff *);
	enum pme_status pm_status;
	unsigned long fd_flags;
	struct sk_buff *skb;

	__pme_stat_inc(TOTAL_NOTIFS);
	sk_token->resp_fd = *fd;

	if (unlikely(in_irq())) {
		__pme_stat_inc(SCAN_IN_IRQ);
		spin_lock_irqsave(&defer_work[coreid].list_lock, flags);
		list_add_tail(&sk_token->list, &defer_work[coreid].list);
		tasklet_schedule(&defer_work[coreid].completion_tasklet);
		spin_unlock_irqrestore(&defer_work[coreid].list_lock,
					flags);
	} else {
		okfn = sk_token->okfn;
		skb = sk_token->skb;

		pm_status = pme_fd_res_status(&sk_token->resp_fd);
		fd_flags = pme_fd_res_flags(&sk_token->resp_fd);

		if (unlikely(pm_status != pme_status_ok))
			pr_err("PME error status 0x%x\n", pm_status);
		/* Is truncation */
		if (unlikely(fd_flags & PME_STATUS_TRUNCATED))
			__pme_stat_inc(NUM_TRUNCATES);
		atomic_inc(&tokens_outstanding);
		kfree(sk_token);
		okfn(skb);
	}
#ifdef DQRR_CONSUME
	return qman_cb_dqrr_consume;
#endif
}

static void scan_ern_cb(struct pme_ctx *ctx, const struct qm_mr_entry *mr,
		struct pme_ctx_token *token)
{
	struct skbuf_token* sk_token = (struct skbuf_token*)token;
	unsigned int coreid = smp_processor_id();
	unsigned long flags;
	int (*okfn)(struct sk_buff *);
	struct sk_buff *skb;

	__pme_stat_inc(NUM_ERNS);
	sk_token->ern = 1;

	if (in_irq()) {
		__pme_stat_inc(SCAN_IN_IRQ);
		spin_lock_irqsave(&defer_work[coreid].list_lock, flags);
		list_add_tail(&sk_token->list, &defer_work[coreid].list);
		tasklet_schedule(&defer_work[coreid].completion_tasklet);
		spin_unlock_irqrestore(&defer_work[coreid].list_lock,
					flags);
	} else {
		okfn = sk_token->okfn;
		skb = sk_token->skb;
		atomic_inc(&tokens_outstanding);
		kfree(sk_token);
		okfn(skb);
	}
}

static int cell_index(struct device_node *node)
{
	const uint32_t	*uint32_prop;
	int lenp;

	uint32_prop = (typeof(uint32_prop))of_get_property(
		node, "cell-index", &lenp);
	if (uint32_prop == NULL) {
		pr_err("Can't get cell-index failed\n");
		return -1;
	}
	return *uint32_prop;
}

static int parent_cell_index(struct device_node *node)
{
	return cell_index(of_get_parent(node));
}

static struct device_node *get_mac_node(struct device_node *dpa_eth_node)
{
	return of_parse_phandle(dpa_eth_node, "fsl,fman-mac", 0);
}

static int eth_chan_id(struct device_node *dpa_eth_node)
{
	struct device_node *chan_node;
	const uint32_t	*uint32_prop;
	int lenp;

	chan_node = of_parse_phandle(dpa_eth_node, "fsl,qman-channel", 0);
	if (chan_node == NULL) {
		pr_err("Can't find chan_node\n");
		return -1;
	} else {
		uint32_prop = (typeof(uint32_prop))
			of_get_property(chan_node, "fsl,qman-channel-id",
					&lenp);
		if (uint32_prop == NULL) {
			pr_err("Get qman-channel-id failed\n");
			return -1;
		} else {
			BUG_ON(lenp != sizeof(uint32_t));
			return *uint32_prop;
		}
	}
}

static int mac_speed(struct device_node *mac_node)
{
	const char *char_prop;
	/* Get the PHY connection type */
	char_prop = (typeof(char_prop))of_get_property(
		mac_node, "phy-connection-type", NULL);
	if (char_prop == NULL) {
		pr_err("Can't get phy-connection-type\n");
		return 0;
	}
	if (strcmp(char_prop, "sgmii") == 0)
		return 1000;
	else if (strcmp(char_prop, "rgmii") == 0)
		return 1000;
	else if (strcmp(char_prop, "xgmii") == 0)
		return 10000;
	pr_err("Unknown speed %s\n", char_prop);
	return -1;
}

static __init int net_scan_demo_init(void)
{
	unsigned i, j;
	int pme_flags = 0;
	int ret;
	struct pme_flow flow;
	struct ctrl_op *ctrl_token = NULL;
	struct proc_dir_entry *entry;
	int total_ctx_init = 0;
	struct qm_fqd_stashing stash_setting;
	struct device_node *dn, *mac_node;
	int dpa_eths = 0;

	pr_info("Loading net_scan_demo module\n");

	atomic_set(&tokens_outstanding, allocs);
	pr_info("Set tokens_outstanding to %d\n", allocs);

	for_each_compatible_node(dn, NULL, "fsl,dpa-ethernet") {
		mac_node =  get_mac_node(dn);
		if (mac_node == NULL)
			pr_err("Can't get fman-mac\n");
		else {
			pr_info("Fman #%d, Eth%d ChanId %d, Speed %d\n",
				parent_cell_index(mac_node)+1,
				cell_index(mac_node),
				eth_chan_id(dn),
				mac_speed(mac_node));
			qman_channel_config[dpa_eths] = eth_chan_id(dn);
		}
		dpa_eths++;
	}
	memset(&stash_setting, 0, sizeof(struct qm_fqd_stashing));
	stash_setting.exclusive = stash_exc;
	stash_setting.annotation_cl = stash_annotation_cl;
	stash_setting.data_cl = stash_data_cl;
	stash_setting.context_cl = stash_ctx_cl;;

	/* Use maximum number of cores */
	if (!num_cores)
		num_cores = num_online_cpus();

	if (!flow_mode)
		num_ctx_per_if = 1;

	if (num_online_cpus() < num_cores) {
		pr_err("Num %u cores wanted is less than num online %d\n",
			num_cores, num_online_cpus());
		return -1;
	}

	if (num_ctx_per_if > MAX_CONTEXT_PER_IF) {
		pr_err("num_ctx_per_if > MAX_CONTEXT_PER_IF %d\n",
			num_ctx_per_if);
		return -1;
	}

	/* How many interfaces are operational, and there mapping */
	pr_info("fm1_mask = 0x%x\n", fm1_mask);
	pr_info("fm2_mask = 0x%x\n", fm2_mask);
	for (i = 0; i < 5; i++) {
		if (fm1_mask & (1<<i)) {
			if (i != 4)
				pr_info("fm1-gb%d ON\n", i);
			else
				pr_info("fm1-10gb ON\n");
			interface_active[i] = 1;
			interface_mapping[i] = num_interfaces;
			interface_being_used[num_interfaces] = i;
			num_interfaces++;
		}
	}
	for (i = 0; i < 5; i++) {
		if (fm2_mask & (1<<i)) {
			if (i != 4)
				pr_info("fm2-gb%d ON\n", i);
			else
				pr_info("fm2-10gb ON\n");
			interface_active[i+5] = 1;
			interface_mapping[i+5] = num_interfaces;
			interface_being_used[num_interfaces] = i+5;
			num_interfaces++;
		}
	}
	pr_info("There are %d Gb/10Gb eth devices used for this module\n",
		 num_interfaces);

	/* Clear stats */
	get_stats(1);
	memset(hw_stats, 0, sizeof(hw_stats));
	clear_all_pme_stats();
	start_time = jiffies;

	for_each_online_cpu(i) {
		active_cores++;
		INIT_LIST_HEAD(&defer_work[i].list);
		spin_lock_init(&defer_work[i].list_lock);
		pr_info("Task init for %d\n",i);
		tasklet_init(&defer_work[i].completion_tasklet,
			process_scanned_tokens,
			(unsigned long)&defer_work[i]);
	}
	entry = create_proc_entry("net_scan_demo_stats", 0, NULL);
	if (entry)
		entry->proc_fops = &proc_ops;
	pme_sw_flow_init(&flow);
	flow.esee = 0;
	flow.ren = residue;
	ctrl_token = kmalloc(sizeof(*ctrl_token), GFP_KERNEL);
	init_completion(&ctrl_token->cb_done);
	ctrl_token->ctx_ctr.cb = ctrl_cb;

	if (!use_orp)
		;
	if (!flow_mode)
		pme_flags |= PME_CTX_FLAG_DIRECT;

	/* Initialize all pme context */
	for (i = 0; i < num_interfaces; i++) {
		for (j = 0; j < num_ctx_per_if; j++) {
			ctx_array[i][j].cb = scan_cb;
			ctx_array[i][j].ern_cb = scan_ern_cb;
			ret = pme_ctx_init(&ctx_array[i][j],
				pme_flags,
				0, /* bpid*/
				qos_in, /* qosin */
				qos_out, /* qosout */
				qman_channel_config[interface_being_used[i]]
				/* stashing */
				, (use_stash ? &stash_setting : NULL));
			if (ret) {
				pr_err("pme_ctx_init failed %d, i=%d,"
					"j=%d\n", ret, i, j);
				goto cleanup_ctx;
			}
			total_ctx_init++;
			ret = pme_ctx_enable(&ctx_array[i][j]);
			if (ret) {
				pr_err("pme_ctx_enable failed %d, i=%d,"
					"j=%d\n", ret, i, j);
				return -1;
			}
			if (flow_mode) {
				ret = pme_ctx_ctrl_update_flow(
					&ctx_array[i][j],
					PME_CTX_OP_WAIT | PME_CMD_FCW_ALL,
					&flow, &ctrl_token->ctx_ctr);
				/* Need a timeout version */
				wait_for_completion(&ctrl_token->cb_done);
			}
			pr_info("Ctx enabled, i=%d, j=%d on qman_chan %d\n"
				, i, j, qman_channel_config[
				interface_being_used[i]]);
		}
	}
	kfree(ctrl_token);
	/* Setup the processing of packets */
	nf_register_hook(&myhook_ops);
	pr_info("Finish net_scan_demo_init\n");
	return 0;

cleanup_ctx:
	for (i = 0; i < num_interfaces; i++) {
		for (j = 0; j < num_ctx_per_if; j++) {
			if (!total_ctx_init)
				goto done;
			ret = pme_ctx_disable(&ctx_array[i][j],
					PME_CTX_OP_WAIT,
					&ctrl_token->ctx_ctr);
			if (ret == 1) {
				wait_for_completion(&ctrl_token->cb_done);
				ret = 0;
			}
			if (ret)
				pr_err("pme_ctx_disable failed %d, i=%d,"
					"j=%d\n", ret, i, j);
			else
				pme_ctx_finish(&ctx_array[i][j]);
			total_ctx_init--;
		}
	}
	remove_proc_entry("net_scan_demo_stats", NULL);
	kfree(ctrl_token);
done:
	return -1;
}

static __exit void net_scan_demo_shutdown(void)
{
	int i, j, ret;
	u64 percent;
	unsigned long end_time;
	unsigned int cpu;
	struct ctrl_op *ctrl_token = NULL;
	u64 counters[NR_PME_EVENTS];

	all_pme_stats(counters);

	ctrl_token = kmalloc(sizeof(*ctrl_token), GFP_KERNEL);
	init_completion(&ctrl_token->cb_done);
	ctrl_token->ctx_ctr.cb = ctrl_cb;

	nf_unregister_hook(&myhook_ops);

	for_each_online_cpu(cpu)
		tasklet_disable(&defer_work[cpu].completion_tasklet);
	end_time = jiffies;

	pr_info("Unloading net_scan_demo module\n");
	pr_info("Processed %llu packets\n"
			"\t(%llu IP packets)"
			"\t (%llu TCP/UDP packets)\n", counters[RX_PACKETS],
			counters[IP_PACKETS], counters[DATA_PACKETS]);
	pr_info("%llu packets matched (%llu total matches)\n",
			counters[MATCHES], counters[TOTAL_MATCHES]);
	percent = counters[DATA_PACKETS] - counters[FULL_FIFO];
	do_div (percent, (end_time - start_time)/HZ);
	pr_info("Test ran for %lu msecs, %llu packets per sec\n",
		   ((end_time - start_time) * 1000)/HZ, percent);
	if (counters[DATA_PACKETS]) {
		percent = counters[FULL_FIFO] * 100;
		do_div(percent, counters[DATA_PACKETS]);
		pr_info("%llu weren't scanned because the FIFO was full"
			" (%llu%%)\n", counters[FULL_FIFO], percent);
	}
	/* Initialize all pme context */
	for (i = 0; i < num_interfaces; i++) {
		for (j = 0; j < num_ctx_per_if; j++) {
			ret = pme_ctx_disable(&ctx_array[i][j],
				PME_CTX_OP_WAIT,  &ctrl_token->ctx_ctr);
			if (ret == 1) {
				wait_for_completion(&ctrl_token->cb_done);
				ret = 0;
			}
			if (ret)
				pr_err("pme_ctx_disableable failed %d,"
					"i=%d, j=%d\n", ret, i, j);
			else
				pme_ctx_finish(&ctx_array[i][j]);
		}
	}
	remove_proc_entry("net_scan_demo_stats", NULL);
	kfree(ctrl_token);
}

/* Module load/unload handlers */
module_init(net_scan_demo_init);
module_exit(net_scan_demo_shutdown);

